@@ -47,11 +47,9 @@ impl ChunkedBuilder<bool, BooleanType> for BooleanChunkedBuilder {
     fn finish(mut self) -> BooleanChunked {
         let arr = Arc::new(self.array_builder.finish());
 
-        let len = arr.len();
         ChunkedArray {
             field: Arc::new(self.field),
             chunks: vec![arr],
-            chunk_id: vec![len],
             phantom: PhantomData,
             categorical_map: None,
         }
@@ -96,11 +94,9 @@ where
     fn finish(mut self) -> ChunkedArray<T> {
         let arr = Arc::new(self.array_builder.finish());
 
-        let len = arr.len();
         ChunkedArray {
             field: Arc::new(self.field),
             chunks: vec![arr],
-            chunk_id: vec![len],
             phantom: PhantomData,
             categorical_map: None,
         }
@@ -162,11 +158,9 @@ impl Utf8ChunkedBuilder {
 
     pub fn finish(mut self) -> Utf8Chunked {
         let arr = Arc::new(self.builder.finish());
-        let len = arr.len();
         ChunkedArray {
             field: Arc::new(self.field),
             chunks: vec![arr],
-            chunk_id: vec![len],
             phantom: PhantomData,
             categorical_map: None,
         }
@@ -313,7 +307,6 @@ where
         ChunkedArray {
             field,
             chunks: vec![Arc::new(builder.finish())],
-            chunk_id: vec![v.len()],
             phantom: PhantomData,
             categorical_map: None,
         }
@@ -367,11 +360,9 @@ where
 macro_rules! finish_list_builder {
     ($self:ident) => {{
         let arr = Arc::new($self.builder.finish());
-        let len = arr.len();
         ListChunked {
             field: Arc::new($self.field.clone()),
             chunks: vec![arr],
-            chunk_id: vec![len],
             phantom: PhantomData,
             categorical_map: None,
         }
@@ -594,7 +585,7 @@ pub fn get_list_builder(
 #[cfg(test)]
 mod test {
     use super::*;
-    use crate::toggle_string_cache;
+    use crate::{reset_string_cache, toggle_string_cache};
 
     #[test]
     fn test_primitive_builder() {
@@ -648,7 +639,9 @@ mod test {
 
     #[test]
     fn test_categorical_builder() {
+        let _lock = crate::SINGLE_LOCK.lock();
         for b in &[false, true] {
+            reset_string_cache();
             toggle_string_cache(*b);
 
             // Use 2 builders to check if the global string cache
