@@ -220,6 +220,17 @@ where
         for ca in other {
             assert_eq!(self.len(), ca.len());
         }
+        if other.len() != (reverse.len() - 1) {
+            return Err(PolarsError::ValueError(
+                format!(
+                    "The amount of ordering booleans: {} does not match that no. of Series: {}",
+                    reverse.len(),
+                    other.len() + 1
+                )
+                .into(),
+            ));
+        }
+
         assert_eq!(other.len(), reverse.len() - 1);
         let mut count: u32 = 0;
         let mut vals: Vec<_> = self
@@ -269,12 +280,33 @@ where
                                 let ca = s.utf8().unwrap();
                                 partial_ord_by_idx!(ca, *reverse)
                             }
-                            _ => {
-                                let ca = self
-                                    .unpack_series_matching_type(s)
-                                    .expect("should be same type");
+                            DataType::Float32 => {
+                                let ca = s.f32().unwrap();
                                 partial_ord_by_idx!(ca, *reverse)
                             }
+                            DataType::Float64 => {
+                                let ca = s.f64().unwrap();
+                                partial_ord_by_idx!(ca, *reverse)
+                            }
+                            DataType::Int64 => {
+                                let ca = s.i64().unwrap();
+                                partial_ord_by_idx!(ca, *reverse)
+                            }
+                            DataType::Int32 => {
+                                let ca = s.i32().unwrap();
+                                partial_ord_by_idx!(ca, *reverse)
+                            }
+                            DataType::UInt32 => {
+                                let ca = s.u32().unwrap();
+                                partial_ord_by_idx!(ca, *reverse)
+                            }
+                            DataType::UInt64 => {
+                                let ca = s.u64().unwrap();
+                                partial_ord_by_idx!(ca, *reverse)
+                            }
+                            _ => {
+                                unreachable!()
+                            }
                         }
                     }
                     // all arrays exhausted, ordering equal it is.
@@ -395,10 +427,33 @@ impl ChunkSort<Utf8Type> for Utf8Chunked {
                                 let ca = s.utf8().unwrap();
                                 partial_ord_by_idx!(ca, *reverse)
                             }
-                            _ => {
-                                let ca = s.f64().expect("cast to f64 before calling this method");
+                            DataType::Float32 => {
+                                let ca = s.f32().unwrap();
                                 partial_ord_by_idx!(ca, *reverse)
                             }
+                            DataType::Float64 => {
+                                let ca = s.f64().unwrap();
+                                partial_ord_by_idx!(ca, *reverse)
+                            }
+                            DataType::Int64 => {
+                                let ca = s.i64().unwrap();
+                                partial_ord_by_idx!(ca, *reverse)
+                            }
+                            DataType::Int32 => {
+                                let ca = s.i32().unwrap();
+                                partial_ord_by_idx!(ca, *reverse)
+                            }
+                            DataType::UInt32 => {
+                                let ca = s.u32().unwrap();
+                                partial_ord_by_idx!(ca, *reverse)
+                            }
+                            DataType::UInt64 => {
+                                let ca = s.u64().unwrap();
+                                partial_ord_by_idx!(ca, *reverse)
+                            }
+                            _ => {
+                                unreachable!()
+                            }
                         }
                     }
                     // all arrays exhausted, ordering equal it is.
@@ -473,6 +528,34 @@ impl ChunkSort<BooleanType> for BooleanChunked {
     }
 }
 
+pub(crate) fn prepare_argsort(
+    columns: Vec<Series>,
+    mut reverse: Vec<bool>,
+) -> Result<(Series, Vec<Series>, Vec<bool>)> {
+    let n_cols = columns.len();
+
+    let mut columns = columns
+        .iter()
+        .map(|s| {
+            use DataType::*;
+            match s.dtype() {
+                Float32 | Float64 | Int32 | Int64 | Utf8 | UInt32 | UInt64 => s.clone(),
+                _ => s.cast::<Int32Type>().unwrap(),
+            }
+        })
+        .collect::<Vec<_>>();
+
+    let first = columns.remove(0);
+
+    // broadcast ordering
+    if n_cols > reverse.len() && reverse.len() == 1 {
+        while n_cols != reverse.len() {
+            reverse.push(reverse[0]);
+        }
+    }
+    Ok((first, columns, reverse))
+}
+
 #[cfg(test)]
 mod test {
     use crate::prelude::*;
