@@ -68,6 +68,7 @@ mod temporal {
         v
     }
 }
+use std::borrow::Cow;
 #[cfg(not(feature = "temporal"))]
 use temporal::*;
 
@@ -92,8 +93,8 @@ macro_rules! format_utf8_array {
                 write!($f, "\tnull\n").ok();
             }
             Some(s) => {
-                if s.len() >= $limit {
-                    write!($f, "\t\"{}...\"\n", &s[..$limit]).ok();
+                if s.len() > 12 {
+                    write!($f, "\t\"{}...\"\n", &s[..12]).ok();
                 } else {
                     write!($f, "\t\"{}\"\n", &s).ok();
                 }
@@ -126,34 +127,19 @@ fn format_object_array(
     name: &str,
     array_type: &str,
 ) -> fmt::Result {
-    write![f, "{}: '{}' [object]\n[\n", array_type, name]?;
+    match object.dtype() {
+        DataType::Object(inner_type) => {
+            write![f, "{}: '{}' [o][{}]\n[\n", array_type, name, inner_type]?;
 
-    for i in 0..limit {
-        let v = object.get(i);
-        match v {
-            AnyValue::Null => writeln!(f, "\tnull")?,
-            _ => writeln!(f, "\tobject")?,
-        }
-    }
-
-    write![f, "]"]
-}
-
-#[cfg(feature = "object")]
-macro_rules! format_object_array {
-    ($limit:expr, $f:expr, $object:expr, $name:expr, $array_type: expr) => {{
-        write![$f, "{}: '{}' [object]\n[\n", $array_type, $name]?;
-
-        for i in 0..$limit {
-            let v = $object.get_any_value(i);
-            match v {
-                AnyValue::Null => writeln!($f, "\tnull")?,
-                _ => writeln!($f, "\tobject")?,
+            for i in 0..limit {
+                let v = object.str_value(i);
+                writeln!(f, "\t{}", v)?;
             }
-        }
 
-        write![$f, "]"]
-    }};
+            write![f, "]"]
+        }
+        _ => unreachable!(),
+    }
 }
 
 macro_rules! set_limit {
@@ -209,7 +195,22 @@ where
 {
     fn fmt(&self, f: &mut Formatter<'_>) -> fmt::Result {
         let limit = set_limit!(self);
-        format_object_array!(limit, f, self, self.name(), "ChunkedArray")
+
+        let taker = self.take_rand();
+        let inner_type = T::type_name();
+        write![
+            f,
+            "ChunkedArray: '{}' [o][{}]\n[\n",
+            self.name(),
+            inner_type
+        ]?;
+        for i in 0..limit {
+            match taker.get(i) {
+                None => writeln!(f, "\tnull")?,
+                Some(val) => writeln!(f, "\t{}", val)?,
+            };
+        }
+        Ok(())
     }
 }
 
@@ -331,17 +332,13 @@ impl Debug for DataFrame {
     }
 }
 
-fn prepare_row(row: Vec<AnyValue>, n_first: usize, n_last: usize) -> Vec<String> {
-    fn make_str_val(v: &AnyValue) -> String {
+fn prepare_row(row: Vec<Cow<'_, str>>, n_first: usize, n_last: usize) -> Vec<String> {
+    fn make_str_val(v: &str) -> String {
         let string_limit = 32;
-        if let AnyValue::Utf8(s) = v {
-            if s.len() > string_limit {
-                format!("\"{}...\"", &s[..string_limit])
-            } else {
-                format!("\"{}\"", s)
-            }
+        if v.len() > string_limit {
+            format!("{}...", &v[..string_limit])
         } else {
-            format!("{}", v)
+            v.to_string()
         }
     }
 
@@ -415,13 +412,13 @@ impl Display for DataFrame {
             let mut rows = Vec::with_capacity(max_n_rows);
             if self.height() > max_n_rows {
                 for i in 0..(max_n_rows / 2) {
-                    let row = self.get(i).unwrap();
+                    let row = self.columns.iter().map(|s| s.str_value(i)).collect();
                     rows.push(prepare_row(row, n_first, n_last));
                 }
                 let dots = rows[0].iter().map(|_| "...".to_string()).collect();
                 rows.push(dots);
                 for i in (self.height() - max_n_rows / 2 - 1)..self.height() {
-                    let row = self.get(i).unwrap();
+                    let row = self.columns.iter().map(|s| s.str_value(i)).collect();
                     rows.push(prepare_row(row, n_first, n_last));
                 }
                 for row in rows {
@@ -429,8 +426,8 @@ impl Display for DataFrame {
                 }
             } else {
                 for i in 0..max_n_rows {
-                    let opt = self.get(i);
-                    if let Some(row) = opt {
+                    if i < self.height() && self.width() > 0 {
+                        let row = self.columns.iter().map(|s| s.str_value(i)).collect();
                         table.add_row(prepare_row(row, n_first, n_last));
                     } else {
                         break;
@@ -456,13 +453,13 @@ impl Display for DataFrame {
             let mut rows = Vec::with_capacity(max_n_rows);
             if self.height() > max_n_rows {
                 for i in 0..(max_n_rows / 2) {
-                    let row = self.get(i).unwrap();
+                    let row = self.columns.iter().map(|s| s.str_value(i)).collect();
                     rows.push(prepare_row(row, n_first, n_last));
                 }
                 let dots = rows[0].iter().map(|_| "...".to_string()).collect();
                 rows.push(dots);
                 for i in (self.height() - max_n_rows / 2 - 1)..self.height() {
-                    let row = self.get(i).unwrap();
+                    let row = self.columns.iter().map(|s| s.str_value(i)).collect();
                     rows.push(prepare_row(row, n_first, n_last));
                 }
                 for row in rows {
@@ -470,8 +467,8 @@ impl Display for DataFrame {
                 }
             } else {
                 for i in 0..max_n_rows {
-                    let opt = self.get(i);
-                    if let Some(row) = opt {
+                    if i < self.height() && self.width() > 0 {
+                        let row = self.columns.iter().map(|s| s.str_value(i)).collect();
                         table.add_row(Row::new(
                             prepare_row(row, n_first, n_last)
                                 .into_iter()
