@@ -1,20 +1,19 @@
+use self::hashing::*;
 use crate::chunked_array::builder::PrimitiveChunkedBuilder;
 use crate::frame::select::Selection;
 use crate::prelude::*;
-use crate::utils::{accumulate_dataframes_vertical, split_ca, split_df, NoNull};
-use crate::vector_hasher::{
-    create_hash_and_keys_threaded_vectorized, df_rows_to_hashes, df_rows_to_hashes_threaded,
-    prepare_hashed_relation, this_thread, IdBuildHasher, IdxHash,
-};
+use crate::utils::{accumulate_dataframes_vertical, split_ca, NoNull};
+use crate::vector_hasher::{AsU64, StrHash};
 use crate::POOL;
 use ahash::RandomState;
-use hashbrown::{hash_map::RawEntryMut, HashMap};
-use itertools::Itertools;
+use hashbrown::HashMap;
+use num::NumCast;
 use rayon::prelude::*;
 use std::fmt::Debug;
-use std::hash::{BuildHasher, Hash};
+use std::hash::{BuildHasher, Hash, Hasher};
 
 pub mod aggregations;
+pub(crate) mod hashing;
 #[cfg(feature = "pivot")]
 pub(crate) mod pivot;
 #[cfg(feature = "downsample")]
@@ -23,237 +22,6 @@ pub mod resample;
 pub type GroupTuples = Vec<(u32, Vec<u32>)>;
 pub type GroupedMap<T> = HashMap<T, Vec<u32>, RandomState>;
 
-fn groupby<T>(a: impl Iterator<Item = T>) -> GroupTuples
-where
-    T: Hash + Eq,
-{
-    let hash_tbl = prepare_hashed_relation(a);
-
-    hash_tbl
-        .into_iter()
-        .map(|(_, indexes)| {
-            let first = unsafe { *indexes.get_unchecked(0) };
-            (first, indexes)
-        })
-        .collect()
-}
-
-fn groupby_threaded_flat<I, T>(iters: Vec<I>, group_size_hint: usize) -> GroupTuples
-where
-    I: IntoIterator<Item = T> + Send,
-    T: Send + Hash + Eq + Sync + Copy,
-{
-    groupby_threaded(iters, group_size_hint)
-        .into_iter()
-        .flatten()
-        .collect()
-}
-
-/// Determine groupby tuples from an iterator. The group_size_hint is used to pre-allocate the group vectors.
-/// When the grouping column is a categorical type we already have a good indication of the avg size of the groups.
-fn groupby_threaded<I, T>(iters: Vec<I>, group_size_hint: usize) -> Vec<GroupTuples>
-where
-    I: IntoIterator<Item = T> + Send,
-    T: Send + Hash + Eq + Sync + Copy,
-{
-    let n_threads = iters.len();
-    let (hashes_and_keys, random_state) = create_hash_and_keys_threaded_vectorized(iters, None);
-    let size = hashes_and_keys.iter().fold(0, |acc, v| acc + v.len());
-
-    // We will create a hashtable in every thread.
-    // We use the hash to partition the keys to the matching hashtable.
-    // Every thread traverses all keys/hashes and ignores the ones that doesn't fall in that partition.
-    POOL.install(|| {
-        (0..n_threads).into_par_iter().map(|thread_no| {
-            let random_state = random_state.clone();
-            let hashes_and_keys = &hashes_and_keys;
-            let thread_no = thread_no as u64;
-
-            let mut hash_tbl: HashMap<T, (u32, Vec<u32>), RandomState> =
-                HashMap::with_capacity_and_hasher(size / n_threads, random_state);
-
-            let n_threads = n_threads as u64;
-            let mut offset = 0;
-            for hashes_and_keys in hashes_and_keys {
-                let len = hashes_and_keys.len() as u32;
-                hashes_and_keys
-                    .iter()
-                    .enumerate()
-                    .for_each(|(idx, (h, k))| {
-                        let idx = idx as u32;
-                        // partition hashes by thread no.
-                        // So only a part of the hashes go to this hashmap
-                        if (h + thread_no) % n_threads == 0 {
-                            let idx = idx + offset;
-                            let entry = hash_tbl
-                                .raw_entry_mut()
-                                // uses the key to check equality to find and entry
-                                .from_key_hashed_nocheck(*h, &k);
-
-                            match entry {
-                                RawEntryMut::Vacant(entry) => {
-                                    let mut tuples = Vec::with_capacity(group_size_hint);
-                                    tuples.push(idx);
-                                    entry.insert_hashed_nocheck(*h, *k, (idx, tuples));
-                                }
-                                RawEntryMut::Occupied(mut entry) => {
-                                    let (_k, v) = entry.get_key_value_mut();
-                                    v.1.push(idx);
-                                }
-                            }
-                        }
-                    });
-
-                offset += len;
-            }
-            hash_tbl.into_iter().map(|(_k, v)| v).collect::<Vec<_>>()
-        })
-    })
-    .collect()
-}
-
-/// Utility function used as comparison function in the hashmap.
-/// The rationale is that equality is an AND operation and therefore its probability of success
-/// declines rapidly with the number of keys. Instead of first copying an entire row from both
-/// sides and then do the comparison, we do the comparison value by value catching early failures
-/// eagerly.
-///
-/// # Safety
-/// Doesn't check any bounds
-pub(crate) unsafe fn compare_df_rows(keys: &DataFrame, idx_a: usize, idx_b: usize) -> bool {
-    for s in keys.get_columns() {
-        if !s.equal_element(idx_a, idx_b, s) {
-            return false;
-        }
-    }
-    true
-}
-
-/// Populate a multiple key hashmap with row indexes.
-/// Instead of the keys (which could be very large), the row indexes are stored.
-/// To check if a row is equal the original DataFrame is also passed as ref.
-/// When a hash collision occurs the indexes are ptrs to the rows and the rows are compared
-/// on equality.
-pub(crate) fn populate_multiple_key_hashmap<V, H, F, G>(
-    hash_tbl: &mut HashMap<IdxHash, V, H>,
-    // row index
-    idx: u32,
-    // hash
-    h: u64,
-    // keys of the hash table (will not be inserted, the indexes will be used)
-    // the keys are needed for the equality check
-    keys: &DataFrame,
-    // value to insert
-    vacant_fn: G,
-    // function that gets a mutable ref to the occupied value in the hash table
-    occupied_fn: F,
-) where
-    G: Fn() -> V,
-    F: Fn(&mut V),
-    H: BuildHasher,
-{
-    let entry = hash_tbl
-        .raw_entry_mut()
-        // uses the idx to probe rows in the original DataFrame with keys
-        // to check equality to find an entry
-        .from_hash(h, |idx_hash| {
-            let key_idx = idx_hash.idx;
-            // Safety:
-            // indices in a groupby operation are always in bounds.
-            unsafe { compare_df_rows(keys, key_idx as usize, idx as usize) }
-        });
-    match entry {
-        RawEntryMut::Vacant(entry) => {
-            entry.insert_hashed_nocheck(h, IdxHash::new(idx, h), vacant_fn());
-        }
-        RawEntryMut::Occupied(mut entry) => {
-            let (_k, v) = entry.get_key_value_mut();
-            occupied_fn(v);
-        }
-    }
-}
-
-fn groupby_multiple_keys(keys: DataFrame) -> GroupTuples {
-    let (hashes, _) = df_rows_to_hashes(&keys, None);
-    let size = hashes.len();
-    // rather over allocate because rehashing is expensive
-    let mut hash_tbl: HashMap<IdxHash, (u32, Vec<u32>), IdBuildHasher> =
-        HashMap::with_capacity_and_hasher(size, IdBuildHasher::default());
-
-    // hashes has no nulls
-    let mut idx = 0;
-    for hashes_chunk in hashes.data_views() {
-        for &h in hashes_chunk {
-            populate_multiple_key_hashmap(
-                &mut hash_tbl,
-                idx,
-                h,
-                &keys,
-                || (idx, vec![idx]),
-                |v| v.1.push(idx),
-            );
-            idx += 1;
-        }
-    }
-    hash_tbl.into_iter().map(|(_k, v)| v).collect::<Vec<_>>()
-}
-
-fn groupby_threaded_multiple_keys_flat(keys: DataFrame, n_threads: usize) -> GroupTuples {
-    let dfs = split_df(&keys, n_threads).unwrap();
-    let (hashes, _random_state) = df_rows_to_hashes_threaded(&dfs, None);
-    let size = hashes.len();
-
-    // We will create a hashtable in every thread.
-    // We use the hash to partition the keys to the matching hashtable.
-    // Every thread traverses all keys/hashes and ignores the ones that doesn't fall in that partition.
-
-    // We use a combination of a custom IdentityHasher and a utility key IdxHash that stores
-    // the index of the row and and the hash. The Hash function of this key just returns the hash it stores.
-    POOL.install(|| {
-        (0..n_threads).into_par_iter().map(|thread_no| {
-            let hashes = &hashes;
-            let thread_no = thread_no as u64;
-
-            let keys = &keys;
-
-            // rather over allocate because rehashing is expensive
-            let mut hash_tbl: HashMap<IdxHash, (u32, Vec<u32>), IdBuildHasher> =
-                HashMap::with_capacity_and_hasher(size / n_threads, IdBuildHasher::default());
-
-            let n_threads = n_threads as u64;
-            let mut offset = 0;
-            for hashes in hashes {
-                let len = hashes.len() as u32;
-
-                let mut idx = 0;
-                for hashes_chunk in hashes.data_views() {
-                    for &h in hashes_chunk {
-                        // partition hashes by thread no.
-                        // So only a part of the hashes go to this hashmap
-                        if this_thread(h, thread_no, n_threads) {
-                            let idx = idx + offset;
-                            populate_multiple_key_hashmap(
-                                &mut hash_tbl,
-                                idx,
-                                h,
-                                &keys,
-                                || (idx, vec![idx]),
-                                |v| v.1.push(idx),
-                            );
-                        }
-                        idx += 1;
-                    }
-                }
-
-                offset += len;
-            }
-            hash_tbl.into_iter().map(|(_k, v)| v).collect::<Vec<_>>()
-        })
-    })
-    .flatten()
-    .collect()
-}
-
 /// Used to create the tuples for a groupby operation.
 pub trait IntoGroupTuples {
     /// Create the tuples need for a groupby operation.
@@ -269,92 +37,142 @@ fn group_multithreaded<T>(ca: &ChunkedArray<T>) -> bool {
     ca.len() > 1000
 }
 
-macro_rules! group_tuples {
-    ($ca: expr, $multithreaded: expr) => {{
-        // TODO! choose a splitting len
-        if $multithreaded && group_multithreaded($ca) {
-            let n_threads = num_cpus::get();
-            let splitted = split_ca($ca, n_threads).unwrap();
+fn num_group_tuples<T>(ca: &ChunkedArray<T>, multithreaded: bool) -> GroupTuples
+where
+    T: PolarsIntegerType,
+    T::Native: Hash + Eq + Send + AsU64,
+    Option<T::Native>: AsU64,
+{
+    let group_size_hint = if let Some(m) = &ca.categorical_map {
+        ca.len() / m.len()
+    } else {
+        0
+    };
+    if multithreaded && group_multithreaded(ca) {
+        let n_threads = POOL.current_num_threads();
+        let splitted = split_ca(ca, n_threads).unwrap();
 
-            if $ca.null_count() == 0 {
-                let iters = splitted
+        // use the arrays as iterators
+        if ca.chunks.len() == 1 {
+            if ca.null_count() == 0 {
+                let keys = splitted
                     .iter()
-                    .map(|ca| ca.into_no_null_iter())
-                    .collect_vec();
-                groupby_threaded_flat(iters, 0)
+                    .map(|ca| ca.cont_slice().unwrap())
+                    .collect::<Vec<_>>();
+                groupby_threaded_num(keys, group_size_hint)
             } else {
-                let iters = splitted.iter().map(|ca| ca.into_iter()).collect_vec();
-                groupby_threaded_flat(iters, 0)
+                let keys = splitted
+                    .iter()
+                    .map(|ca| {
+                        ca.downcast_iter()
+                            .map(|v| v.into_iter().map(|v| v.as_u64()))
+                            .flatten()
+                            .collect::<Vec<_>>()
+                    })
+                    .collect::<Vec<_>>();
+
+                groupby_threaded_num(keys, group_size_hint)
             }
+            // use the polars-iterators
+        } else if ca.null_count() == 0 {
+            let keys = splitted
+                .iter()
+                .map(|ca| ca.into_no_null_iter().collect::<Vec<_>>())
+                .collect::<Vec<_>>();
+            groupby_threaded_num(keys, group_size_hint)
         } else {
-            if $ca.null_count() == 0 {
-                groupby($ca.into_no_null_iter())
-            } else {
-                groupby($ca.into_iter())
-            }
+            let keys = splitted
+                .iter()
+                .map(|ca| ca.into_iter().map(|v| v.as_u64()).collect::<Vec<_>>())
+                .collect::<Vec<_>>();
+            groupby_threaded_num(keys, group_size_hint)
         }
-    }};
+    } else if ca.null_count() == 0 {
+        groupby(ca.into_no_null_iter())
+    } else {
+        groupby(ca.into_iter())
+    }
 }
 
 impl<T> IntoGroupTuples for ChunkedArray<T>
 where
-    T: PolarsIntegerType,
-    T::Native: Eq + Hash + Send,
+    T: PolarsNumericType,
+    T::Native: NumCast,
 {
     fn group_tuples(&self, multithreaded: bool) -> GroupTuples {
-        let group_size_hint = if let Some(m) = &self.categorical_map {
-            self.len() / m.len()
-        } else {
-            0
-        };
-        if multithreaded && group_multithreaded(self) {
-            let n_threads = num_cpus::get();
-            let splitted = split_ca(self, n_threads).unwrap();
-
-            // use the arrays as iterators
-            if self.chunks.len() == 1 {
-                if self.null_count() == 0 {
-                    let iters = splitted
-                        .iter()
-                        .map(|ca| ca.downcast_iter().map(|array| array.values()))
-                        .flatten()
-                        .collect_vec();
-                    groupby_threaded_flat(iters, group_size_hint)
-                } else {
-                    let iters = splitted
-                        .iter()
-                        .map(|ca| ca.downcast_iter())
-                        .flatten()
-                        .collect_vec();
-                    groupby_threaded_flat(iters, group_size_hint)
-                }
-                // use the polars-iterators
-            } else if self.null_count() == 0 {
-                let iters = splitted
-                    .iter()
-                    .map(|ca| ca.into_no_null_iter())
-                    .collect_vec();
-                groupby_threaded_flat(iters, group_size_hint)
-            } else {
-                let iters = splitted.iter().map(|ca| ca.into_iter()).collect_vec();
-                groupby_threaded_flat(iters, group_size_hint)
+        match self.dtype() {
+            DataType::UInt64 => {
+                // convince the compiler that we are this type.
+                let ca: &UInt64Chunked = unsafe {
+                    &*(self as *const ChunkedArray<T> as *const ChunkedArray<UInt64Type>)
+                };
+                num_group_tuples(ca, multithreaded)
+            }
+            DataType::UInt32 => {
+                // convince the compiler that we are this type.
+                let ca: &UInt32Chunked = unsafe {
+                    &*(self as *const ChunkedArray<T> as *const ChunkedArray<UInt32Type>)
+                };
+                num_group_tuples(ca, multithreaded)
+            }
+            DataType::Int64 | DataType::Float64 => {
+                let ca = self.bit_repr_large();
+                num_group_tuples(&ca, multithreaded)
+            }
+            DataType::Int32 | DataType::Float32 => {
+                let ca = self.bit_repr_small();
+                num_group_tuples(&ca, multithreaded)
+            }
+            _ => {
+                let ca = self.cast::<UInt32Type>().unwrap();
+                num_group_tuples(&ca, multithreaded)
             }
-        } else if self.null_count() == 0 {
-            groupby(self.into_no_null_iter())
-        } else {
-            groupby(self.into_iter())
         }
     }
 }
 impl IntoGroupTuples for BooleanChunked {
     fn group_tuples(&self, multithreaded: bool) -> GroupTuples {
-        group_tuples!(self, multithreaded)
+        self.cast::<UInt32Type>()
+            .unwrap()
+            .group_tuples(multithreaded)
     }
 }
 
 impl IntoGroupTuples for Utf8Chunked {
     fn group_tuples(&self, multithreaded: bool) -> GroupTuples {
-        group_tuples!(self, multithreaded)
+        let hb = RandomState::default();
+        if multithreaded {
+            let n_threads = POOL.current_num_threads();
+            let splitted = split_ca(self, n_threads).unwrap();
+
+            let str_hashes = POOL.install(|| {
+                splitted
+                    .par_iter()
+                    .map(|ca| {
+                        ca.into_iter()
+                            .map(|opt_s| {
+                                let mut state = hb.build_hasher();
+                                opt_s.hash(&mut state);
+                                let hash = state.finish();
+                                StrHash::new(opt_s, hash)
+                            })
+                            .collect::<Vec<_>>()
+                    })
+                    .collect::<Vec<_>>()
+            });
+            groupby_threaded_num(str_hashes, 0)
+        } else {
+            let str_hashes = self
+                .into_iter()
+                .map(|opt_s| {
+                    let mut state = hb.build_hasher();
+                    opt_s.hash(&mut state);
+                    let hash = state.finish();
+                    StrHash::new(opt_s, hash)
+                })
+                .collect::<Vec<_>>();
+            groupby(str_hashes.iter())
+        }
     }
 }
 
@@ -366,46 +184,6 @@ impl IntoGroupTuples for CategoricalChunked {
     }
 }
 
-macro_rules! impl_into_group_tpls_float {
-    ($self: ident, $multithreaded:expr) => {
-        if $multithreaded && group_multithreaded($self) {
-            let n_threads = num_cpus::get();
-            let splitted = split_ca($self, n_threads).unwrap();
-            match $self.null_count() {
-                0 => {
-                    let iters = splitted
-                        .iter()
-                        .map(|ca| ca.into_no_null_iter().map(|v| v.to_bits()))
-                        .collect_vec();
-                    groupby_threaded_flat(iters, 0)
-                }
-                _ => {
-                    let iters = splitted
-                        .iter()
-                        .map(|ca| ca.into_iter().map(|opt_v| opt_v.map(|v| v.to_bits())))
-                        .collect_vec();
-                    groupby_threaded_flat(iters, 0)
-                }
-            }
-        } else {
-            match $self.null_count() {
-                0 => groupby($self.into_no_null_iter().map(|v| v.to_bits())),
-                _ => groupby($self.into_iter().map(|opt_v| opt_v.map(|v| v.to_bits()))),
-            }
-        }
-    };
-}
-
-impl IntoGroupTuples for Float64Chunked {
-    fn group_tuples(&self, multithreaded: bool) -> GroupTuples {
-        impl_into_group_tpls_float!(self, multithreaded)
-    }
-}
-impl IntoGroupTuples for Float32Chunked {
-    fn group_tuples(&self, multithreaded: bool) -> GroupTuples {
-        impl_into_group_tpls_float!(self, multithreaded)
-    }
-}
 impl IntoGroupTuples for ListChunked {}
 #[cfg(feature = "object")]
 impl<T> IntoGroupTuples for ObjectChunked<T> {}
@@ -1077,7 +855,7 @@ impl<'df, 'selection_str> GroupBy<'df, 'selection_str> {
         Column: AsRef<str>,
     {
         // create a mapping from columns to aggregations on that column
-        let mut map = HashMap::with_capacity_and_hasher(column_to_agg.len(), RandomState::new());
+        let mut map = HashMap::with_hasher(RandomState::new());
         column_to_agg.iter().for_each(|(column, aggregations)| {
             map.insert(column.as_ref(), aggregations.as_ref());
         });
@@ -1253,7 +1031,7 @@ pub fn fmt_groupby_column(name: &str, method: GroupByMethod) -> String {
 mod test {
     use itertools::Itertools;
 
-    use crate::frame::groupby::{groupby, groupby_threaded_flat};
+    use crate::frame::groupby::{groupby, groupby_threaded_num};
     use crate::prelude::*;
     use crate::utils::split_ca;
     use num::traits::FloatConst;
@@ -1511,11 +1289,13 @@ mod test {
             vec![1, 2, 3, 4, 4, 4, 2, 1, 1],
             vec![1, 2, 3, 4, 4, 4],
         ] {
-            let ca = UInt8Chunked::new_from_slice("", &slice);
+            let ca = UInt32Chunked::new_from_slice("", &slice);
             let splitted = split_ca(&ca, 4).unwrap();
 
             let a = groupby(ca.into_iter()).into_iter().sorted().collect_vec();
-            let b = groupby_threaded_flat(splitted.iter().map(|ca| ca.into_iter()).collect(), 0)
+
+            let keys = splitted.iter().map(|ca| ca.cont_slice().unwrap()).collect();
+            let b = groupby_threaded_num(keys, 0)
                 .into_iter()
                 .sorted()
                 .collect_vec();
