@@ -1,12 +1,25 @@
 #[allow(clippy::wrong_self_convention)]
+/// Allow selection by:
+///
+/// &str => df.select("my-column"),
+/// (&str)" => df.select(("col_1", "col_2")),
+/// Vec<&str)" => df.select(vec!["col_a", "col_b"]),
 pub trait Selection<'a, S> {
     fn to_selection_vec(self) -> Vec<&'a str>;
+
+    // a single column is selected
+    fn single(&self) -> Option<&str> {
+        None
+    }
 }
 
 impl<'a> Selection<'a, &str> for &'a str {
     fn to_selection_vec(self) -> Vec<&'a str> {
         vec![self]
     }
+    fn single(&self) -> Option<&'a str> {
+        Some(self)
+    }
 }
 
 impl<'a> Selection<'a, &str> for Vec<&'a str> {
@@ -23,6 +36,14 @@ where
     fn to_selection_vec(self) -> Vec<&'a str> {
         self.as_ref().iter().map(|s| s.as_ref()).collect()
     }
+
+    fn single(&self) -> Option<&str> {
+        let a = self.as_ref();
+        match a.len() {
+            1 => Some(a[0].as_ref()),
+            _ => None,
+        }
+    }
 }
 
 impl<'a> Selection<'a, &str> for (&'a str, &'a str) {
