@@ -33,6 +33,7 @@ pub(crate) mod private {
     use crate::frame::groupby::GroupTuples;
 
     use ahash::RandomState;
+    use std::borrow::Cow;
 
     pub trait PrivateSeries {
         unsafe fn equal_element(
@@ -151,6 +152,10 @@ pub(crate) mod private {
                 "argsort_multiple is not implemented for this Series".into(),
             ))
         }
+        /// Formatted string representation. Can used in formatting.
+        fn str_value(&self, _index: usize) -> Cow<str> {
+            unimplemented!()
+        }
     }
 }
 
@@ -890,10 +895,25 @@ pub trait SeriesTrait: Send + Sync + private::PrivateSeries {
     #[cfg(feature = "temporal")]
     #[cfg_attr(docsrs, doc(cfg(feature = "temporal")))]
     /// Format Date32/Date64 with a `fmt` rule. See [chrono strftime/strptime](https://docs.rs/chrono/0.4.19/chrono/format/strftime/index.html).
-    fn datetime_str_fmt(&self, fmt: &str) -> Result<Series> {
+    fn strftime(&self, fmt: &str) -> Result<Series> {
         match self.dtype() {
-            DataType::Date32 => self.date32().map(|ca| ca.str_fmt(fmt).into_series()),
-            DataType::Date64 => self.date64().map(|ca| ca.str_fmt(fmt).into_series()),
+            DataType::Date32 => self.date32().map(|ca| ca.strftime(fmt).into_series()),
+            DataType::Date64 => self.date64().map(|ca| ca.strftime(fmt).into_series()),
+            _ => Err(PolarsError::InvalidOperation(
+                format!("operation not supported on dtype {:?}", self.dtype()).into(),
+            )),
+        }
+    }
+
+    #[cfg(feature = "temporal")]
+    #[cfg_attr(docsrs, doc(cfg(feature = "temporal")))]
+    /// Convert date(time) object to timestamp in ms.
+    fn timestamp(&self) -> Result<Int64Chunked> {
+        match self.dtype() {
+            DataType::Date32 => self
+                .date32()
+                .map(|ca| (ca.cast::<Int64Type>().unwrap() * 1000)),
+            DataType::Date64 => self.date64().map(|ca| ca.cast::<Int64Type>().unwrap()),
             _ => Err(PolarsError::InvalidOperation(
                 format!("operation not supported on dtype {:?}", self.dtype()).into(),
             )),
