@@ -84,6 +84,40 @@ impl VecHash for Float64Chunked {
 
 impl VecHash for ListChunked {}
 
+// Used to to get a u64 from the hashing keys
+pub(crate) trait AsU64 {
+    #[allow(clippy::wrong_self_convention)]
+    fn as_u64(self) -> u64;
+}
+
+impl AsU64 for u32 {
+    fn as_u64(self) -> u64 {
+        self as u64
+    }
+}
+
+impl AsU64 for u64 {
+    fn as_u64(self) -> u64 {
+        self
+    }
+}
+
+impl AsU64 for Option<u32> {
+    fn as_u64(self) -> u64 {
+        match self {
+            Some(v) => v as u64,
+            // just a number
+            None => 13,
+        }
+    }
+}
+
+impl AsU64 for Option<u64> {
+    fn as_u64(self) -> u64 {
+        self.unwrap_or(13)
+    }
+}
+
 pub struct IdHasher {
     hash: u64,
 }
@@ -127,6 +161,9 @@ impl Default for IdHasher {
 pub type IdBuildHasher = BuildHasherDefault<IdHasher>;
 
 #[derive(Debug)]
+/// Contains an idx of a row in a DataFrame and the precomputed hash of that row.
+/// That hash still needs to be used to create another hash to be able to resize hashmaps without
+/// accidental quadratic behavior. So do not use an Identity function!
 pub(crate) struct IdxHash {
     // idx in row of Series, DataFrame
     pub(crate) idx: u32,
@@ -147,66 +184,45 @@ impl IdxHash {
     }
 }
 
+/// Contains a ptr to the string slice an the precomputed hash of that string.
+/// During rehashes, we will rehash the hash instead of the string, that makes rehashing
+/// cheap and allows cache coherent small hash tables.
+#[derive(Eq, Copy, Clone)]
+pub(crate) struct StrHash<'a> {
+    str: Option<&'a str>,
+    hash: u64,
+}
+
+impl<'a> Hash for StrHash<'a> {
+    fn hash<H: Hasher>(&self, state: &mut H) {
+        state.write_u64(self.hash)
+    }
+}
+
+impl<'a> StrHash<'a> {
+    pub(crate) fn new(s: Option<&'a str>, hash: u64) -> Self {
+        Self { str: s, hash }
+    }
+}
+
+impl<'a> PartialEq for StrHash<'a> {
+    fn eq(&self, other: &Self) -> bool {
+        self.str == other.str
+    }
+}
+
+impl<'a> AsU64 for StrHash<'a> {
+    fn as_u64(self) -> u64 {
+        self.hash
+    }
+}
+
 /// Check if a hash should be processed in that thread.
 #[inline]
 pub(crate) fn this_thread(h: u64, thread_no: u64, n_threads: u64) -> bool {
     (h + thread_no) % n_threads == 0
 }
 
-fn finish_table_from_key_hashes<T>(
-    hashes_nd_keys: Vec<(u64, T)>,
-    mut hash_tbl: HashMap<T, Vec<u32>, RandomState>,
-    offset: usize,
-) -> HashMap<T, Vec<u32>, RandomState>
-where
-    T: Hash + Eq,
-{
-    hashes_nd_keys
-        .into_iter()
-        .enumerate()
-        .for_each(|(idx, (h, t))| {
-            let idx = (idx + offset) as u32;
-
-            let entry = hash_tbl
-                .raw_entry_mut()
-                // uses the key to check equality to find and entry
-                .from_key_hashed_nocheck(h, &t);
-
-            match entry {
-                RawEntryMut::Vacant(entry) => {
-                    entry.insert_hashed_nocheck(h, t, vec![idx]);
-                }
-                RawEntryMut::Occupied(mut entry) => {
-                    let (_k, v) = entry.get_key_value_mut();
-                    v.push(idx);
-                }
-            }
-        });
-    hash_tbl
-}
-
-pub(crate) fn prepare_hashed_relation<T>(
-    b: impl Iterator<Item = T>,
-) -> HashMap<T, Vec<u32>, RandomState>
-where
-    T: Hash + Eq,
-{
-    let build_hasher = RandomState::default();
-
-    let hashes_nd_keys = b
-        .map(|val| {
-            let mut hasher = build_hasher.build_hasher();
-            val.hash(&mut hasher);
-            (hasher.finish(), val)
-        })
-        .collect::<Vec<_>>();
-
-    let hash_tbl: HashMap<T, Vec<u32>, RandomState> =
-        HashMap::with_capacity_and_hasher(hashes_nd_keys.len(), build_hasher);
-
-    finish_table_from_key_hashes(hashes_nd_keys, hash_tbl, 0)
-}
-
 pub(crate) fn prepare_hashed_relation_threaded<T, I>(
     iters: Vec<I>,
 ) -> Vec<HashMap<T, Vec<u32>, RandomState>>
@@ -216,7 +232,6 @@ where
 {
     let n_threads = iters.len();
     let (hashes_and_keys, build_hasher) = create_hash_and_keys_threaded_vectorized(iters, None);
-    let size = hashes_and_keys.iter().fold(0, |acc, v| acc + v.len());
 
     // We will create a hashtable in every thread.
     // We use the hash to partition the keys to the matching hashtable.
@@ -227,7 +242,7 @@ where
             let hashes_and_keys = &hashes_and_keys;
             let thread_no = thread_no as u64;
             let mut hash_tbl: HashMap<T, Vec<u32>, RandomState> =
-                HashMap::with_capacity_and_hasher(size / (5 * n_threads), build_hasher);
+                HashMap::with_hasher(build_hasher);
 
             let n_threads = n_threads as u64;
             let mut offset = 0;
